/*
 * wtmpclean.c -- clear from a wtmp database the unwanted records
 * Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * indent -i4 -nut src/wtmpclean.c
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <errno.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>

/* FIXME */
/*#if !defined HAVE_UTMPXPOSIX
 *# include "wrappers.h"
 *#endif
 */

#ifdef HAVE_UTMPX_H
# include <utmpx.h>
#else
# ifdef HAVE_UTMP_H
#  include <utmp.h>
# endif
#endif

#include "wtmpclean.h"
#include "getopt.h"

static char *progname;

static unsigned char opt_list = 0;
static unsigned char opt_preserve = 0;
static unsigned char opt_raw = 0;

void
usage (int status)
{
    static const char *usagemsg[] = {
        PACKAGE " version " PACKAGE_VERSION,
        "Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>",
        "",
#if defined(DEBUG)
        "Usage: " PACKAGE " [-l|-r] [-p] [ -t YYYYMMDDHHMM ]"
#else
        "Usage: " PACKAGE " [-l] [-p] [ -t YYYYMMDDHHMM ]"
#endif
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
            " [-f <wtmpfile>]"
#endif
            " <username>",
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "  -f, --file       Modify <wtmpfile> instead of " DEFAULT_WTMP,
#endif
        "  -l, --list       Show listing of <user> logins",
        "  -p, --preserve   Preserve ownership and timestamps",
#if defined(DEBUG)
        "  -r, --raw        Show the raw content of the wtmp database",
#endif
        "  -t, --time       Delete the login at the specified time",
        "",
        "Samples:",
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "  ./" PACKAGE " root -t 200809061430",
        "  ./" PACKAGE " -p -f " DEFAULT_WTMP ".1 root",
#else
        "  ./" PACKAGE " --preserve root",
#endif
        "",
        "This is free software; see the source for copying conditions.  There is NO",
        "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.",
        (char *) 0
    };

    unsigned int i;

    for (i = 0; usagemsg[i]; i++)
        fprintf (status ? stderr : stdout, "%s\n", usagemsg[i]);

    exit (status);
}

void
die (char *msg, int err)
{
    printf ("%s: %s: %s\n", basename (progname), msg, strerror (err));
    exit (1);
}

void
listentry (struct utmpxlist *p, int what)
{
    char *ct;

    printf ("%s   ", p->ut.ut_user);
    printf ("%-28.28s  ", p->ut.ut_line);
    ct = ctime (&p->ut.ut_tv.tv_sec);
    printf ("%10.10s %4.4s %5.5s", ct, ct + 20, ct + 11);

    switch (what)
      {
      case R_REBOOT:
          printf ("  ");
          break;
      case R_NOW:
          printf (" - still logged in");
          break;
      case R_PHANTOM:
          printf ("   gone - no logout");
          break;
      case R_NORMAL:
      default:
          ct = ctime (&p->eos);
          printf (" - %5.5s  ", ct + 11);
          if (p->delta < SECDAY)
              printf ("(%5.5s)  ", asctime (gmtime (&p->delta)) + 11);
          else
              printf ("(%ld+%5.5s)  ", p->delta / SECDAY,
                      asctime (gmtime (&p->delta)) + 11);
      }

    printf ("%s\n", p->ut.ut_host);
}

unsigned int
wtmpedit (const char *wtmpfile, const char *user, unsigned int preserve,
          struct tm *logintime, unsigned int *cleanerr)
{
    static struct utmpx *utp;
    unsigned int cleanrec;
    struct stat sb;
    struct utimbuf currtime;
    uid_t owner = 0;
    gid_t group = 0;
    time_t tm;

    if (access (wtmpfile, W_OK))
        die ("cannot access the file", errno);

    if (preserve)
      {
          if (stat (wtmpfile, &sb))
              die ("cannot get file status", errno);

          currtime.actime = sb.st_atime;
          currtime.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;
      }

    UTMPXNAME (wtmpfile);
    setutxent ();

    tm = mktime (logintime);
/*  if ((tm = mktime(&u)) == (time_t)-1)
 *        ...
 */

    cleanrec = *cleanerr = 0;
    while ((utp = getutxent ()))
      {
          if (utp->ut_type == USER_PROCESS &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                if ((tm != (time_t) - 1) &&
                    !((utp->ut_tv.tv_sec - (utp->ut_tv.tv_sec % 60) <= tm)
                      && (tm <
                          utp->ut_tv.tv_sec - (utp->ut_tv.tv_sec % 60) + 59)))
                    continue;

                utp->ut_type = DEAD_PROCESS;
                memset (utp->ut_line, 0, sizeof utp->ut_line);
                utp->ut_tv.tv_sec = utp->ut_tv.tv_usec = 0;
                memset (utp->ut_user, 0, sizeof utp->ut_user);
                memset (utp->ut_host, 0, sizeof utp->ut_host);  /* ??? */
#ifdef HAVE_UTP_UT_ADDR_V6
                memset (utp->ut_addr_v6, 0, sizeof utp->ut_addr_v6);    /* ??? */
#endif
                /* ... utp->ut_pid ... */

                if (pututxline (utp))
                    cleanrec++;
                else
                    cleanerr++;
            }
      }

    endutxent ();

    if (preserve && chown (wtmpfile, owner, group) < 0)
      {
          fprintf (stderr, "cannot preserve the ownership of the wtmp file");
      }
    if (preserve && utime (wtmpfile, &currtime) < 0)
      {
          fprintf (stderr, "cannot preserve access and modification times");
      }

    return cleanrec;
}

void
wtmpxlist (const char *wtmpfile, const char *user)
{
    struct utmpxlist *p, *curr = NULL, *next, *utmpxlist = NULL;
    struct utmpx *utp;
    char *ct;
    int ltype;

    if (access (wtmpfile, R_OK))
        die ("cannot access the file", errno);

    UTMPXNAME (wtmpfile);
    setutxent ();

    while ((utp = getutxent ()))
      {
          if (opt_raw)
            {
                if (strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
                  {
                      printf ("%s   ", utp->ut_user);
                      printf ("%-28.28s  ", utp->ut_line);
                      ct = ctime (&utp->ut_tv.tv_sec);
                      printf ("%10.10s %4.4s %5.5s\n", ct, ct + 20, ct + 11);
                  }
                continue;
            }

          switch (utp->ut_type)
            {
            case USER_PROCESS:
                /*
                 * Just store the data if it is interesting enough.
                 */
                if (strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
                  {
                      if ((p = malloc (sizeof (struct utmpxlist))) == NULL)
                          die ("out of memory", errno);

                      memcpy (&p->ut, utp, sizeof (struct utmpx));
                      p->delta = -1;    /* '-1' means "unset" */
                      p->next = NULL;
                      if (utmpxlist == NULL)
                        {
                            utmpxlist = curr = p;
                            p->prev = NULL;
                        }
                      else
                        {
                            curr->next = p;
                            p->prev = curr;
                            curr = p;
                        }
                  }
                break;
            case BOOT_TIME:
                /* listentry (utp, R_REBOOT); */
                break;
            case DEAD_PROCESS:
                for (p = curr; p; p = p->prev)
                  {
                      if (p->ut.ut_type != USER_PROCESS || p->delta != -1)
                          continue;
                      /*if (strncmp (p->ut.ut_user, utp->ut_user, sizeof utp->ut_user))
                       *    continue;
                       */
                      if (strncmp
                          (p->ut.ut_line, utp->ut_line, sizeof utp->ut_line))
                          continue;

                      p->eos = utp->ut_tv.tv_sec;
                      p->delta = utp->ut_tv.tv_sec - p->ut.ut_tv.tv_sec;

                      if (utp->ut_type == DEAD_PROCESS)
                          break;
                  }
                break;
            case INIT_PROCESS:
            case LOGIN_PROCESS:
            case OLD_TIME:
            case NEW_TIME:
                /*case SHUTDOWN_TIME: */
                break;
            }
      }

    for (p = utmpxlist; p; p = next)
      {
          next = p->next;

          ltype = R_NORMAL;
          if (p->delta < 0)
            {
                /* Is process still alive? */
                if (p->ut.ut_pid > 0 && kill (p->ut.ut_pid, 0) != 0
                    && errno == ESRCH)
                    ltype = R_PHANTOM;
                else
                    ltype = R_NOW;
            }

          listentry (p, ltype);
          free (p);
      }

    endutxent ();
}

int
main (int argc, char **argv)
{
    const char *wtmpfile = DEFAULT_WTMP;
    int opt, opt_index = 0;
    unsigned int cleanrec, cleanerr;
    char *user = NULL;
    struct tm u = { 0 };

    static struct option long_options[] = {
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        {"file", required_argument, 0, 'f'},
#endif
        {"list", no_argument, 0, 'l'},
        {"preserve", no_argument, 0, 'p'},
#if defined(DEBUG)
        {"raw", no_argument, 0, 'r'},
#endif
        {"time", required_argument, 0, 't'},
        {"help", no_argument, 0, 'h'},
        {0, 0, 0, 0}
    };
    static const char *options =
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "f:"
#endif
#if defined(DEBUG)
        "r"
#endif
        "lpt:h";

    opterr = 0;
    progname = argv[0];

    while (1)
      {
          opt = getopt_long (argc, argv, options, long_options, &opt_index);
          if (opt == -1)
              break;

          switch (opt)
            {
            case 'f':
                wtmpfile = optarg;
                break;
            case 'l':
                opt_list = 1;
                break;
            case 'p':
                opt_preserve = 1;
                break;
            case 'r':
                opt_raw = 1;
                break;
            case 't':
                if (sscanf (optarg, "%4d%2d%2d%2d%2d", &u.tm_year,
                            &u.tm_mon, &u.tm_mday, &u.tm_hour,
                            &u.tm_min) != 5)
                  {
                      fprintf (stderr, "%s: invalid time value \"%s\"\n",
                               basename (progname), optarg);
                      exit (1);
                  }

                u.tm_year -= 1900;
                u.tm_mon -= 1;
                u.tm_isdst = -1;
                break;
            default:
                usage ((opt == 'h') ? 1 : 0);
            }
      }

    if (argc != optind + 1)
      {
          usage (1);
      }

    user = argv[optind];

    if (opt_list || opt_raw)
      {
          wtmpxlist (wtmpfile, user);
          exit (0);
      }

    cleanrec = wtmpedit (wtmpfile, user, opt_preserve, &u, &cleanerr);

    if (cleanerr > 0)
      {
          fprintf (stderr, "%s: cannot clean up %s\n", progname, wtmpfile);
      }
    else
      {
          printf ("%u block(s) logging user %s removed from %s\n",
                  cleanrec, user, wtmpfile);
      }

    return 0;
}
